;; Generated package description from ellama.el  -*- no-byte-compile: t -*-
(define-package "ellama" "0.13.11.0.20250209.202508" "Tool for interacting with LLMs" '((emacs "28.1") (llm "0.22.0") (spinner "1.7.4") (transient "0.7") (compat "29.1")) :commit "6fb94007b983a81f06e693569faf70513a1e2157" :authors '(("Sergey Kostyaev" . "sskostyaev@gmail.com")) :maintainer '("Sergey Kostyaev" . "sskostyaev@gmail.com") :keywords '("help" "local" "tools") :url "http://github.com/s-kostyaev/ellama")
